/*
 * Copyright © 2012 Intel Corporation
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 *
 * Authors:
 *    Daniel Vetter <daniel.vetter@ffwll.ch>
 *
 */

#include "igt.h"
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <inttypes.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/ioctl.h>

#include "drm.h"
#include "i915/gem.h"

IGT_TEST_DESCRIPTION("Check set_tiling vs pwrite coherency.");

#define OBJECT_SIZE (1024*1024)
#define TEST_STRIDE (1024*4)

/**
 * Testcase: Check set_tiling vs pwrite coherency
 */

igt_simple_main
{
	uint32_t data[OBJECT_SIZE/4];
	uint32_t *ptr;
	uint32_t handle;
	int fd;

	fd = drm_open_driver(DRIVER_INTEL);
	igt_require(gem_available_fences(fd) > 0);
	gem_require_pread_pwrite(fd);

	for (int i = 0; i < OBJECT_SIZE/4; i++)
		data[i] = i;

	handle = gem_create(fd, OBJECT_SIZE);
	ptr = __gem_mmap__gtt(fd, handle, OBJECT_SIZE, PROT_WRITE);

	gem_set_tiling(fd, handle, I915_TILING_X, TEST_STRIDE);

	/* touch it */
	if (ptr) {
		gem_set_domain(fd, handle,
			       I915_GEM_DOMAIN_GTT, I915_GEM_DOMAIN_GTT);
		*ptr = 0xdeadbeef;
	}

	igt_info("testing pwrite on tiled buffer\n");
	gem_write(fd, handle, 0, data, OBJECT_SIZE);
	memset(data, 0, OBJECT_SIZE);
	gem_read(fd, handle, 0, data, OBJECT_SIZE);
	for (int i = 0; i < OBJECT_SIZE/4; i++)
		igt_assert_eq_u32(data[i], i);

	/* touch it before changing the tiling, so that the fence sticks around */
	if (ptr) {
		gem_set_domain(fd, handle,
			       I915_GEM_DOMAIN_GTT, I915_GEM_DOMAIN_GTT);
		*ptr = 0xdeadbeef;
	}

	gem_set_tiling(fd, handle, I915_TILING_NONE, 0);

	igt_info("testing pwrite on untiled, but still fenced buffer\n");
	gem_write(fd, handle, 0, data, OBJECT_SIZE);
	memset(data, 0, OBJECT_SIZE);
	gem_read(fd, handle, 0, data, OBJECT_SIZE);
	for (int i = 0; i < OBJECT_SIZE/4; i++)
		igt_assert_eq_u32(data[i], i);

	munmap(ptr, OBJECT_SIZE);

	close(fd);
}
