/*
 * Copyright (C) 2019 Intel Corporation
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#ifndef I915_PERF_RECORDER_COMMANDS_H
#define I915_PERF_RECORDER_COMMANDS_H

#include <stdint.h>

#define I915_PERF_RECORD_FIFO_PATH "/tmp/.i915-perf-record"

enum recorder_command {
	RECORDER_COMMAND_DUMP = 1,
	RECORDER_COMMAND_QUIT,
};

struct recorder_command_base {
	uint32_t command;
	uint32_t size; /* size of recorder_command_base + dump in bytes */
};

/*
 The dump after the recorder_command_base header:

struct recorder_command_dump {
	uint8_t path[0];
};
*/

#endif
