/*
 * Copyright © 2015 Intel Corporation
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

#ifndef IGT_H
#define IGT_H

#include "drmtest.h"
#include "i915_3d.h"
#include "i915_pciids.h"
#include "igt_aux.h"
#include "igt_core.h"
#include "igt_core.h"
#include "igt_debugfs.h"
#include "igt_draw.h"
#include "igt_dummyload.h"
#include "igt_fb.h"
#include "igt_frame.h"
#include "igt_gt.h"
#include "igt_kms.h"
#include "igt_params.h"
#include "igt_pm.h"
#include "igt_stats.h"
#ifdef HAVE_CHAMELIUM
#include "igt_alsa.h"
#include "igt_audio.h"
#include "igt_chamelium.h"
#include "igt_chamelium_stream.h"
#endif
#include "instdone.h"
#include "intel_batchbuffer.h"
#include "intel_chipset.h"
#include "intel_io.h"
#include "ioctl_wrappers.h"
#include "media_fill.h"
#include "media_spin.h"
#include "rendercopy.h"
#include "i915/gem_mman.h"
#include "i915/gem_engine_topology.h"

#endif /* IGT_H */
